import json
import threading

import cotea.utils as cotea_utils

# during the imports ansible global objects are created
# that affects to the further work
cotea_utils.remove_modules_from_imported(module_name_like="ansible", not_to_delete="ansible_runner")
cotea_utils.remove_modules_from_imported(module_name_like="logging")

from ansible.cli import CLI
from ansible.plugins.strategy.linear import StrategyModule
from ansible.plugins.strategy import StrategyBase
from ansible.cli.playbook import PlaybookCLI
from ansible.parsing.yaml.objects import AnsibleUnicode
from ansible.executor.playbook_executor import PlaybookExecutor
from ansible.executor.play_iterator import PlayIterator
from ansible import constants as C
from ansible.playbook.task import Task
from ansible.playbook.helpers import load_list_of_tasks

from cotea.ansible_execution_sync import ans_sync
from cotea.task_result import TaskResult
from cotea.wrappers.pbcli_run_wrapper import pbcli_run_wrapper
from cotea.wrappers.strategy_run_wrapper import strategy_run_wrapper
from cotea.wrappers.get_next_task_wrapper import get_next_task_wrapper
from cotea.wrappers.update_active_conn_wrapper import update_active_conn_wrapper
from cotea.wrappers.play_prereqs_wrapper import play_prereqs_wrapper
from cotea.wrappers.playbook_executor_wrapper import play_executor_wrapper
from cotea.wrappers.iterator_add_task_wrapper import iterator_add_task_wrapper
from cotea.progress_bar import ansible_progress_bar
from cotea.ansible_execution_tree import AnsibleExecTree

import logging


class runner:
    def __init__(self, pb_path, arg_maker, debug_mod=None, show_progress_bar=False):
        logging_lvl = logging.INFO
        if debug_mod:
            logging_lvl= logging.DEBUG

        self.show_progress_bar = show_progress_bar

        logging.basicConfig(format="%(name)s %(asctime)s %(message)s", \
                    datefmt="%H:%M:%S", level=logging_lvl)

        self.pb_path = pb_path
        self.arg_maker = arg_maker

        self.logger = logging.getLogger("RUNNER")

        log_sync = logging.getLogger("SYNC")
        self.sync_obj = ans_sync(log_sync)

        self.breakpoint_labeles = {}
        self.breakpoint_labeles["before_playbook"] = "before_playbook_run"
        self.breakpoint_labeles["after_playbook"] = "after_playbook_run"
        self.breakpoint_labeles["before_play"] = "before_play_run"
        self.breakpoint_labeles["after_play"] = "after_play_run"
        self.breakpoint_labeles["before_task"] = "before_task_run"
        self.breakpoint_labeles["after_task"] = "after_task_run"

        self.progress_bar = ansible_progress_bar()
        self.execution_tree = AnsibleExecTree()

        self._set_wrappers()
        start_ok = self._start_ansible()
        self.logger.debug("Ansible start ok: %s", start_ok)


    def _set_wrappers(self):
        wrp_lgr = logging.getLogger("WRPR")

        self.pbcli_run_wrp = pbcli_run_wrapper(PlaybookCLI.run, self.sync_obj, wrp_lgr,
                                               self.breakpoint_labeles["before_playbook"],
                                               self.breakpoint_labeles["after_playbook"])
        PlaybookCLI.run = self.pbcli_run_wrp

        self.play_wrp = strategy_run_wrapper(StrategyModule.run, self.sync_obj, wrp_lgr,
                                             self.breakpoint_labeles["before_play"],
                                             self.breakpoint_labeles["after_play"])
        StrategyModule.run = self.play_wrp

        self.task_wrp = get_next_task_wrapper(StrategyModule._get_next_task_lockstep,
                                              self.sync_obj, wrp_lgr,
                                              self.breakpoint_labeles["before_task"],
                                              self.progress_bar,
                                              self.show_progress_bar)
        StrategyModule._get_next_task_lockstep = self.task_wrp

        self.update_conn_wrapper = update_active_conn_wrapper(StrategyBase.update_active_connections,
                                                              self.sync_obj, wrp_lgr,
                                                              self.breakpoint_labeles["after_task"])
        StrategyBase.update_active_connections = self.update_conn_wrapper

        self.play_prereqs_wrp = play_prereqs_wrapper(CLI._play_prereqs,
                                                      self.sync_obj, wrp_lgr)
        CLI._play_prereqs = self.play_prereqs_wrp

        if self.show_progress_bar:
            self.playbook_executor_wrp = play_executor_wrapper(PlaybookExecutor.__init__,
                                                            self.sync_obj, wrp_lgr,
                                                            self.execution_tree,
                                                            self.progress_bar)
            PlaybookExecutor.__init__ = self.playbook_executor_wrp

        self.iterator_add_task_wrp = iterator_add_task_wrapper(PlayIterator.add_tasks,
                                                                  self.sync_obj, wrp_lgr,
                                                                  self.execution_tree,
                                                                  self.progress_bar)
        PlayIterator.add_tasks = self.iterator_add_task_wrp


    def _set_wrappers_back(self):
        PlaybookCLI.run = self.pbcli_run_wrp.func
        StrategyModule.run = self.play_wrp.func
        StrategyModule._get_next_task_lockstep = self.task_wrp.func
        StrategyBase.update_active_connections = self.update_conn_wrapper.func
        CLI._play_prereqs = self.play_prereqs_wrp.func
        PlayIterator.add_tasks = self.iterator_add_task_wrp.func
        if self.show_progress_bar:
            PlaybookExecutor.__init__ = self.playbook_executor_wrp.func

    def _except_hook(self, args, /):
        if (args.exc_type == SystemExit or
                # NOTE: this probably should never happen
                args.thread != self.ansible_thread):
            return self._old_except_hook(args)

        self.sync_obj.exception = args.exc_value
        self.sync_obj.continue_runner()

    def _start_ansible(self):
        args = self.arg_maker.args
        args.insert(0, "/usr/local/bin/ansible-playbook")
        args.insert(1, self.pb_path)

        self.pbCLI = PlaybookCLI(args)

        self.ansible_thread = threading.Thread(target=self.pbCLI.run)
        self._old_except_hook = threading.excepthook
        threading.excepthook = self._except_hook

        self.ansible_thread.start()
        self.sync_obj.runner_just_wait()

        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["before_playbook"]:
            return True

        return False


    def has_next_play(self):
        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["after_playbook"]:
            return False

        self.sync_obj.continue_ansible_with_stop()
        current_bp_label = self.sync_obj.curr_breakpoint_label
        self.logger.debug("has_next_play: %s", current_bp_label)

        if current_bp_label == self.breakpoint_labeles["before_play"]:
            return True

        return False


    def has_next_task(self):
        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["after_playbook"]:
            return False

        self.sync_obj.continue_ansible_with_stop()
        current_bp_label = self.sync_obj.curr_breakpoint_label
        self.logger.debug("has_next_task: %s", current_bp_label)

        if current_bp_label == self.breakpoint_labeles["before_task"]:
            return True

        return False


    def run_next_task(self):
        res = []

        if self.sync_obj.curr_breakpoint_label == self.breakpoint_labeles["after_playbook"]:
            return res

        self.sync_obj.continue_ansible_with_stop()
        current_bp_label = self.sync_obj.curr_breakpoint_label
        self.logger.debug("run_next_task: %s", current_bp_label)

        if current_bp_label != self.breakpoint_labeles["after_task"]:
            self.logger.debug("run_next_task() has come not in to the 'after_task'")

        for task_result_ansible_obj in self.update_conn_wrapper.current_results:
            res.append(TaskResult(task_result_ansible_obj))

        self.task_wrp.set_next_to_prev()

        return res


    def rerun_last_task(self):
        self.task_wrp.rerun_last_task = True


    # returns True and empty string if success
    #         False and error msg otherwise
    def add_new_task(self, new_task_str, is_dict=False):
        prev_task = self.get_prev_task()
        curr_block = None

        has_attrs, error_msg = cotea_utils.obj_has_attrs(prev_task, ["_parent"])
        if not has_attrs:
            return False, error_msg

        curr_block = prev_task._parent
        block_attrs = ["_loader", "_play", "_role", "_variable_manager", "_use_handlers"]
        has_attrs, error_msg = cotea_utils.obj_has_attrs(curr_block, block_attrs)
        if not has_attrs:
            return False, error_msg

        loader = curr_block._loader

        ds = None
        if not is_dict:
            try:
                ds = loader.load(new_task_str)
            except Exception as e:
                error_msg = "Exception during loader.load call, is_dict is {} "
                error_msg += "(from str to python ds): {}"
                return False, error_msg.format(is_dict, str(e))
        else:
            try:
                new_task_str_dict = json.loads(new_task_str)
            except Exception as e:
                error_msg = "Exception during json.loads call, is_dict is {} "
                error_msg += "(from str-aka-dict to python ds): {}"
                return False, error_msg.format(is_dict, str(e))
            ds = [new_task_str_dict]

        #print("DS:\n", ds)

        has_attrs, _ = cotea_utils.obj_has_attrs(ds, ["__len__"])
        if not has_attrs:
            error_msg = "Python repr of the input string should have "
            error_msg += "__len__ attr. Maybe something wrong with input: {}\n"
            error_msg += "Python repr without __len__ attr: {}"
            return False, error_msg.format(new_task_str, str(ds))

        if len(ds) != 1:
            error_msg = "You must add 1 new task. Instead you add: {}"
            return False, error_msg.format(str(ds))

        curr_play = curr_block._play
        #curr_role = curr_block._role
        variable_manager = curr_block._variable_manager
        use_handlers=curr_block._use_handlers

        try:
            new_ansible_task = load_list_of_tasks(
                ds=ds,
                play=curr_play,
                block=curr_block,
                #role=curr_role,
                variable_manager=variable_manager,
                loader=loader,
                use_handlers=use_handlers,
            )

        except Exception as e:
            error_msg = "Exception during load_list_of_tasks call "
            error_msg += "(creats Ansible.Task objects): {}"
            return False, error_msg.format(str(e))

        has_attrs, _ = cotea_utils.obj_has_attrs(new_ansible_task, ["__len__"])
        if not has_attrs:
            error_msg = "Python repr of the input string should have "
            error_msg += "__len__ attr. Maybe something wrong with input: {}\n"
            error_msg += "Python repr without __len__ attr: {}"
            return False, error_msg.format(new_task_str, str(ds))

        new_tasks_count = len(new_ansible_task)
        if new_tasks_count != 1:
            error_msg = "The input '{}' has been interpreted into {} tasks "
            error_msg += "instead of 1. Interpretation result: {}"
            return False, error_msg.format(new_task_str, new_tasks_count, str(ds))

        #self.task_wrp.new_task_to_add = True
        self.task_wrp.new_task = new_ansible_task[0]

        adding_res, error_msg = self.task_wrp.add_tasks(new_ansible_task)

        return adding_res, error_msg


    def get_new_added_task(self):
        return self.task_wrp.new_task


    def ignore_errors_of_next_task(self):
        self.task_wrp.next_task_ignore_errors = True


    def dont_add_last_task_after_new(self):
        self.task_wrp.dont_add_last_task_after_new()


    def get_already_ignore_failed(self):
        return self.task_wrp.should_ignored_errors_uuids


    def get_already_ignore_unrch(self):
        return self.task_wrp.should_ignored_unrch_uuids


    def finish_ansible(self):
        if self.sync_obj.exception is None:
            while self.sync_obj.curr_breakpoint_label != self.breakpoint_labeles["after_playbook"]:
                self.sync_obj.continue_ansible_with_stop()
            self.sync_obj.continue_ansible()

        self.ansible_thread.join(timeout=5)
        self._set_wrappers_back()


    def get_cur_play_name(self):
        return str(self.play_wrp.current_play_name)


    def get_next_task(self):
        return self.task_wrp.get_next_task()


    def get_next_task_name(self):
        return str(self.task_wrp.get_next_task_name())


    def get_prev_task(self):
        return self.task_wrp.get_prev_task()


    def get_prev_task_name(self):
        return str(self.task_wrp.get_prev_task_name())


    def get_last_task_result(self):
        res = []

        for task_result_ansible_obj in self.update_conn_wrapper.current_results:
            res.append(TaskResult(task_result_ansible_obj))

        return res


    # returns True if there was an non ignored error
    def was_error(self):
        return self.play_wrp.was_error


    # returns list with all errors, including the ignored ones
    def get_all_error_msgs(self):
        return self.update_conn_wrapper.error_msgs


    # returns last error msg that wasn't ignored 
    def get_error_msg(self):
        res = ""

        # the errors didn't have 'ignore_errors'
        if self.was_error():
            errors_count = len(self.update_conn_wrapper.error_msgs)

            if errors_count > 0:
                res = self.update_conn_wrapper.error_msgs[errors_count - 1]

        return res


    def get_all_vars(self):
        variable_manager = self.play_wrp.variable_manager
        cur_play = self.play_wrp.iterator._play
        hosts = self.play_wrp.hosts
        hosts_all = self.play_wrp.hosts_all

        res = variable_manager.get_vars(play=cur_play,
                                        _hosts=hosts,
                                        _hosts_all=hosts_all)

        return res


    def get_all_facts(self):
        return self.play_prereqs_wrp.variable_manager._fact_cache.copy()


    def get_variable(self, var_name):
        if var_name == "ansible_facts":
            return self.get_all_facts()

        all_vars = self.get_all_vars()

        if var_name in all_vars:
            return all_vars[var_name]

        # check groups
        if "groups" in all_vars:
            if var_name in all_vars["groups"]:
                return all_vars["groups"][var_name]

        result = {}

        # check hostvars
        if "hostvars" in all_vars:
            for host in all_vars["hostvars"]:
                for key in all_vars["hostvars"][host]:
                    if key == var_name:
                        result[host] = {key: all_vars["hostvars"][host][key]}

        if result:
            return result

        facts = self.get_all_facts()
        for host_key in facts:
            if var_name in facts[host_key]:
                result[host_key] = facts[host_key][var_name]

        if result:
            return result

        self.logger.info("There is no variable with name %s", var_name)

        return None


    def add_var_as_extra_var(self, new_var_name, value):
        variable_manager = self.play_wrp.variable_manager

        ansible_way_var = AnsibleUnicode(new_var_name)
        variable_manager._extra_vars[ansible_way_var] = value


    def get_stats(self):
        return self.play_wrp.custom_stats


    def skip_next_task(self):
        skipped_task_name = self.task_wrp.skip_next_task()
        return skipped_task_name


from cotea.wrappers.wrapper_base import wrapper_base


# wraps from ansible.cli.CLI._play_prereqs()
class play_prereqs_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger):
        super().__init__(func, sync_obj, logger)

        self.loader = None
        self.inventory = None
        self.variable_manager = None

    def __call__(self, real_obj):
        loader, inventory, variable_manager = self.func()
        self.logger.debug("_play_prereqs() call")

        self.loader = loader
        self.inventory = inventory
        self.variable_manager = variable_manager

        return loader, inventory, variable_manager



PLAY_NAME_IND = 0
PLAY_TASKS_IND = 1

class AnsibleExecTree:
    def __init__(self):
        self.tree = []
        #self.hosts_tasks = {}
        self.task_count = -1
    

    def add_play(self, play_name):
        play_exist = False
        for play in self.tree:
            if play[PLAY_NAME_IND] == play_name:
                play_exist = True
                break
        
        if not play_exist:
            self.tree.append([play_name, {}])
    

    def add_task(self, play_name, host_name, task_name):
        added = False

        for play in self.tree:
            if play[PLAY_NAME_IND] == play_name:
                if host_name not in play[PLAY_TASKS_IND]:
                    play[PLAY_TASKS_IND][host_name] = []

                play[PLAY_TASKS_IND][host_name].append(task_name)
                added = True

                '''
                if host_name not in self.hosts_tasks:
                    self.hosts_tasks[host_name] = []
                
                self.hosts_tasks[host_name].append(task_name)
                '''

                break
    

    def compute_metrics(self):
        self.task_count = 0

        for play in self.tree:
            tmp_max = -1

            for host in play[PLAY_TASKS_IND]:
                tmp_max = max(tmp_max, len(play[PLAY_TASKS_IND][host]))
            
            self.task_count += tmp_max


    def pretty_print(self):
        for play in self.tree:
            print("Play name:", play[PLAY_NAME_IND])
            
            for host_name in play[PLAY_TASKS_IND]:
                print("Host name:", host_name)
                print(play[PLAY_TASKS_IND][host_name], "\n")
            
            print("\n")
        
        print("Task count -", self.task_count)




class ansible_progress_bar:
    def __init__(self):
        self.executed_count = 0
        self.bar_len = 100
        self.bar_sym = u"█"
        self.bar_empty_sym = "."
    

    def set_total_task_count(self, tasks_count):
        self.total_task_count = tasks_count
    
    def update(self):
        self.executed_count += 1


    def add_to_total_task_count(self, num):
        self.total_task_count += num


    def print_bar(self, play_name, task_name):
        percent = self.executed_count / self.total_task_count
        divisions = int(percent * self.bar_len)
        if divisions > self.bar_len:
            divisions = self.bar_len

        if divisions < 0:
            divisions = 0
        
        divisions_str = self.bar_sym*divisions

        percent = percent * 100
        empty_divisions = self.bar_empty_sym*(self.bar_len - divisions)

        bar = "Next task: {}. Play: {}\n[{}{}]  {} / {}, {:.2f}%".format(task_name,
                                                                       play_name,
                                                                       divisions_str,
                                                                       empty_divisions,
                                                                       self.executed_count,
                                                                       self.total_task_count,
                                                                       percent)
        
        print(bar, "\n")




import tempfile

from ansible.playbook.play_context import PlayContext
from ansible.executor.play_iterator import PlayIterator
from ansible.playbook import Playbook
from ansible.template import Templar

from cotea.wrappers.wrapper_base import wrapper_base
from cotea.ansible_execution_tree import AnsibleExecTree
from cotea.progress_bar import ansible_progress_bar

TASK_IND = 1


# wraps from ansible.executor.playbook_executor.PlaybookExecutor()
class play_executor_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger, exec_tree: AnsibleExecTree,
                 progress_bar: ansible_progress_bar):
        super().__init__(func, sync_obj, logger)

        self.ansible_exec_tree = exec_tree
        self.progress_bar = progress_bar

        self.loader = None
        self.variable_manager = None
        self.inventory = None


    def __call__(self, real_obj, playbooks, inventory, variable_manager, loader, passwords):
        self.loader = loader
        self.variable_manager = variable_manager
        self.inventory = inventory

        self.func(real_obj, playbooks, inventory, variable_manager, loader, passwords)
        connection_lockfile = tempfile.TemporaryFile()

        for playbook in playbooks:
            pb = Playbook.load(playbook, variable_manager, loader)

            plays = pb.get_plays()

            for play in plays:
                new_play = play.copy()
                play_name = play.get_name()
                play_context = PlayContext(new_play, passwords, connection_lockfile)

                all_vars = variable_manager.get_vars(play=play)
                iterator = PlayIterator(
                    inventory=inventory,
                    play=new_play,
                    play_context=play_context,
                    variable_manager=variable_manager,
                    all_vars=all_vars,
                    start_at_done=False,
                )

                hosts_cache = _get_hosts_cache(new_play, inventory)
                hosts_left = _get_hosts_left(inventory, iterator, hosts_cache)

                self.ansible_exec_tree.add_play(play_name)
                
                for host in hosts_left:
                    host_str = str(host)
                    while True:
                        task = iterator.get_next_task_for_host(host)
                        if not task[TASK_IND]:
                            break
                        
                        if hasattr(task[TASK_IND], "get_name"):
                            task_name = task[TASK_IND].get_name()
                            self.ansible_exec_tree.add_task(play_name, host_str, task_name)
                        else:
                            self.logger.debug("playbook executor, task[TASK_IND] has no attr get_name: ", task[TASK_IND])
        
        self.ansible_exec_tree.compute_metrics()
        self.progress_bar.set_total_task_count(self.ansible_exec_tree.task_count)
        #self.ansible_exec_tree.pretty_print()



def _get_hosts_cache(play, inventory):
    if Templar(None).is_template(play.hosts):
        _pattern = 'all'
    else:
        _pattern = play.hosts or 'all'
    
    hosts_cache_all = [h.name for h in inventory.get_hosts(pattern=_pattern, ignore_restrictions=True)]
    hosts_cache = [h.name for h in inventory.get_hosts(play.hosts, order=play.order)]

    return hosts_cache
    

def _get_hosts_left(inventory, iterator, hosts_cache):
    hosts_left = []
    for host in hosts_cache:
        try:
            hosts_left.append(inventory.hosts[host])
        except KeyError:
            hosts_left.append(inventory.get_host(host))
    return hosts_left


from ansible.executor.play_iterator import PlayIterator
from ansible.playbook.task import Task
from ansible.playbook.block import Block

from cotea.wrappers.wrapper_base import wrapper_base
from cotea.progress_bar import ansible_progress_bar
from cotea.ansible_execution_tree import AnsibleExecTree

TASK_IND = 1


# wraps from ansible.executor.play_iterator.PlayIterator.add_tasks()
class iterator_add_task_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger, exec_tree: AnsibleExecTree,
                 progress_bar: ansible_progress_bar):
        super().__init__(func, sync_obj, logger)

        self.ansible_exec_tree = exec_tree
        self.progress_bar = progress_bar


    def __call__(self, real_obj: PlayIterator, host, all_blocks):
        res = self.func(real_obj, host, all_blocks)
        
        play_name = real_obj._play.get_name()
        host_name = str(host)

        self.ansible_exec_tree.add_play(play_name)
        for block in all_blocks:
            block_tasks = _get_all_block_tasks(block)
            for task_name in block_tasks:
                self.ansible_exec_tree.add_task(play_name, host_name, task_name)

        self.ansible_exec_tree.compute_metrics()
        self.progress_bar.set_total_task_count(self.ansible_exec_tree.task_count)

        # except Exception as e:
        #     print("--------Bug find print--------")

        #     print("Exception is: ", e)
        #     print("Play name: ", play_name)
        #     print("Blocks: ", all_blocks)

        #     block_num = 1
        #     for block in all_blocks:
        #         print("Block num =", block_num)
        #         for task in block.block:
        #             print("type of blocks.block elem: ", type(task))
        #             print("blocks.block elem:", task, "\n")

        #             if hasattr(task, "block"):
        #                 print(task.block)
                
        #         block_num += 1

        #     print("------------------------------")

        return res


def _get_all_block_tasks(block):
    new_tasks = []

    _get_all_block_tasks_rec(new_tasks, block)

    return new_tasks


def _get_all_block_tasks_rec(new_tasks: list, block: Block):
    if hasattr(block, "block"):
        for task_or_block in block.block:
            if isinstance(task_or_block, Task):
                if hasattr(task_or_block, "get_name"):
                    #print("\tappending")
                    new_tasks.append(task_or_block.get_name())
                else:
                    print("task has no 'get_name' attr:", task_or_block)
            
            elif isinstance(task_or_block, Block):
                _get_all_block_tasks_rec(new_tasks, task_or_block)
            
            else:
                print("in 'block' list object is not Task/Block type:", type(task_or_block), task_or_block)

    #else:
    #    print("block has no 'block' attr:", block)


from ansible.executor.task_result import TaskResult as TR


class TaskResult:
    def __init__(self, ansible_task_result: TR):
        self.result = ansible_task_result._result.copy()
        self.task_name = str(ansible_task_result.task_name)
        self.task_ansible_object = ansible_task_result._task
        self.task_fields = ansible_task_result._task_fields.copy()

        self.original_task_uuid = ""
        if hasattr(ansible_task_result._task, "_uuid"):
            self.original_task_uuid = ansible_task_result._task._uuid

        self.is_changed = ansible_task_result.is_changed()
        self.is_failed = ansible_task_result.is_failed()
        self.is_skipped = ansible_task_result.is_skipped()
        self.is_unreachable = ansible_task_result.is_unreachable()

        self.stdout = ""
        stdout = self.result.get("stdout")
        if stdout:
            self.stdout = str(stdout)
        
        self.stderr = ""
        stderr = self.result.get("stderr")
        if stderr:
            self.stderr = str(stderr)

        self.msg = ""
        msg = self.result.get("msg")
        if msg:
            self.msg = str(msg)
        else:
            msg = self.result.get("message")
            if msg:
                self.msg = str(msg)


    def pretty_string(self):
        task_info = "&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n"
        task_info += "Task with name {}\n"
        task_info += "is_changed - {}\n"
        task_info += "is_failed - {}\n"
        task_info += "is_skipped - {}\n"
        task_info += "is_unreachable - {}\n"
        task_info += "stdout - {}\n"
        task_info += "stderr - {}\n"
        task_info += "msg - {}\n"
        task_info += "result dict - {}\n"
        task_info += "task fields - {}\n"

        task_info += "\n"

        task_info = task_info.format(
            self.task_name,
            self.is_changed,
            self.is_failed,
            self.is_skipped,
            self.is_unreachable,
            self.stdout,
            self.stderr,
            self.msg,
            str(self.result),
            str(self.task_fields)
        )

        t = self.task_ansible_object
        d = dir(t)
        
        for attr in d:
            task_info += "{} - {}\n".format(attr, getattr(t, attr))

        task_info += "&&&&&&&&&&&&&&&&&&&&&&&&&&&&\n"

        return task_info
        


from ansible.playbook.task import Task
from cotea.runner import runner


def print_help_msg():
    help_msg = "Informative commands:\n"
    help_msg += "'ft' - print info about the Failed Task\n"
    help_msg += "'msg' - print all ansible error MSGs (including the ignored ones)\n"
    help_msg += "'p' - print Progress bar\n"
    help_msg += "'h'/'help' - print this msg\n"
    help_msg += "\nAction commands:\n"
    help_msg += "'re' - RErun of the failed task\n"
    help_msg += "'v' - add new Variable as extra var\n"
    help_msg += "'c' - Continue ansible execution (go to the next task)\n"
    help_msg += "'nt' - add New Task\n"
    help_msg += "'w' - Watch ansible variable value\n"
    #help_msg += "'f' - Finish ansible execution\n"

    print(help_msg)


def pretty_print_task(task: Task):
    pretty_print = "Name: {}\nAction: {}\nArgs: {}\nbecome: {}\n".format(task.name,
                                                                         task.action,
                                                                         task.args,
                                                                         task.become)
    
    
    pretty_print_p2 = "Connection: {}\ndelegate_facts: {}\nenvironment: {}\n".format(task.connection,
                                                                                     task.delegate_facts,
                                                                                     task.environment)

    pretty_print += pretty_print_p2

    pretty_print_p2 = "delegate_facts: {}\nvars: {}\n".format(task.delegate_facts,
                                                              task.get_vars())
    pretty_print += pretty_print_p2

    pretty_print_p2 = "port: {}\nremote_user: {}\nretries: {}\ntags: {}\n".format(task.port,
                                                                                  task.remote_user,
                                                                                  task.retries,
                                                                                  task.tags)
    pretty_print += pretty_print_p2

    print(pretty_print)


def interactive_discotech(failed_task: Task, r: runner):
    print("\nINTERACTIVE MODE")
    
    while True:
        print("Enter command: ", end="")
        command = input()
        if command == "ft":
            pretty_print_task(failed_task)

        elif command == "re":
            r.rerun_last_task()
            r.progress_bar.add_to_total_task_count(1)
            break

        elif command == "msg":
            info_msg = "This msg list includes msg's that were ignored. "
            info_msg += "Msg's will be displayed in order of occurrence.\n"
            print(info_msg)

            err_msgs = r.get_all_error_msgs()
            msg_number = 1
            
            for msg in err_msgs:
                print("MSG number {}:\n{}\n".format(msg_number, msg))
                msg_number += 1

        elif command == "v":
            print("var will be added as extra var")
            print("Enter var name: ", end="")
            var_name = input()
            print("Enter var value: ", end="")
            var_value = input()
                                
            r.add_var_as_extra_var(var_name, var_value)
            print("var added successfully!\n")
        
        elif command == "c":
            break

        elif command == "p":
            print()
            play_name = r.get_cur_play_name()
            next_task = r.get_next_task_name()
            r.progress_bar.print_bar(play_name, next_task)
        
        elif command == "nt":
            print("Enter new task like a string entering all \\n and spaces needed:")
            new_task_str = input()

            # TODO: not sure that this is a good solution
            #       however, this is interactive mode and
            #       we always can say to user that he is wrong
            new_task_str = new_task_str.replace("\\n", "\n")
            add_ok, err_msg = r.add_new_task(new_task_str)

            if not add_ok:
                print("\nThe adding process was failed with the error:\n", err_msg)
            else:
                print("\nNew task was added! It will run on every host of current inventory.")
                print("Press 'c' after this, not 're'. This will run new task and the failed one after it.\n")
                r.progress_bar.add_to_total_task_count(1)
        
        elif command == "w":
            print("Enter var name: ", end="")
            
            var_name = input()
            value = r.get_variable(var_name)
            msg = "{} var value:\n{}".format(var_name, value)
            
            print(msg)

        elif command == "help" or command == "h":
            print_help_msg()

        else:
            print("Enter commands correctly! 'help' or just 'h' will help you\n")



def remove_modules_from_imported(module_name_like, not_to_delete=None):
    import sys

    for module in list(sys.modules):
        if module_name_like in module:
            if not_to_delete is not None:
                if not_to_delete in module:
                    continue
            
            sys.modules.pop(module)

#  True, ""           - if has
#  False, "error msg" - if hasn't
def obj_has_attrs(obj, attrs_list):
    for attr in attrs_list:
        if not hasattr(obj, attr):
            error_msg = "Obj {} doesn't have attr {} but should"
            return False, error_msg.format(obj, attr)
    
    return True, ""


from cotea.wrappers.wrapper_base import wrapper_base
from cotea.wrappers.ansi_breakpoint import ansi_breakpoint


# wraps ansible.plugins.strategy.linear.StrategyModule.run()
class strategy_run_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger, before_bp_label, after_bp_label):
        super().__init__(func, sync_obj, logger)

        self.before_play_bp = ansi_breakpoint(sync_obj, before_bp_label)
        self.after_play_bp = ansi_breakpoint(sync_obj, after_bp_label)
        self.current_play_name = None
        self.iterator = None
        self.play_context = None
        self.variable_manager = None
        self.hosts = None
        self.hosts_all = None
        self.was_error = False
        self.strategy_obj = None
        self.stats = None


    def __call__(self, real_obj, iterator, play_context):
        self.logger.debug("play run")

        self.was_error = False
        self.iterator = iterator
        self.play_context = play_context
        self.variable_manager = real_obj._variable_manager
        self.hosts = real_obj._hosts_cache
        self.hosts_all = real_obj._hosts_cache_all
        self.custom_stats = None
        
        if hasattr(real_obj._tqm, "_stats"):
            if hasattr(real_obj._tqm._stats, "custom"):
                self.custom_stats = real_obj._tqm._stats.custom

        self.strategy_obj = real_obj
        try:
            self.current_play_name = iterator._play.get_name()
        except:
            pass
        
        self.before_play_bp.stop()

        result = self.func(real_obj, iterator, play_context)

        # self.iterator = None
        # self.play_context = None
        # self.current_play_name = None
        
        self.logger.debug("play end")
        
        self.after_play_bp.stop()
        if result != real_obj._tqm.RUN_OK:
            self.was_error = True

        return result
    



class argument_maker:
    def __init__(self):
        self.args = []

    def add_arg(self, param_name, value=None):
        self.args.append(str(param_name))
        if value:
            self.args.append(str(value))


'''
    def get_args(self):
        return help_class(self.options)


class help_class(object):
    def __init__(self, d):
        for a, b in d.items():
            if isinstance(b, (list, tuple)):
               setattr(self, a, [help_class(x) if isinstance(x, dict) else x for x in b])
            else:
               setattr(self, a, help_class(b) if isinstance(b, dict) else b)


default_options = {'listhosts': None, 'ssh_common_args': '', 'step': None, 'ask_vault_pass': False,
    'private_key_file': None, 'diff': False, 'args': ['test_playbook.yml'], 'timeout': 10,
    'extra_vars': [], 'become_method': 'sudo', 'connection': 'smart', 'vault_password_files': [],
    'module_path': None, 'become': False, 'check': False, 'flush_cache': None,
    'inventory': ['/home/steve/vagrant_practice/vagrant_getting_started/hosts'], 'skip_tags': [],
    'tags': ['all'], 'force_handlers': False, 'become_user': None, 'listtags': None, 'vault_ids': [],
    'forks': 5, 'scp_extra_args': '', 'subset': None, 'ask_pass': False, 'verbosity': 3, 'remote_user': None,
    'become_ask_pass': False, 'sftp_extra_args': '', 'start_at_task': None, 'syntax': None, 'listtasks': None,
    'ssh_extra_args': ''}
'''



import functools


class wrapper_base:
    def __init__(self, func, sync_obj, logger):
        functools.update_wrapper(self, func)
        self.func = func
        self.sync_obj = sync_obj
        self.logger = logger

    # for decorating class methods(bound methods)
    def __get__(self, instance, owner):
        return functools.partial(self.__call__, instance)



from cotea.wrappers.wrapper_base import wrapper_base
from cotea.wrappers.ansi_breakpoint import ansi_breakpoint


# wraps ansible.plugins.strategy.StrategyBase.update_active_connections()
class update_active_conn_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger, bp_label):
        super().__init__(func, sync_obj, logger)

        self.after_task_bp = ansi_breakpoint(sync_obj, bp_label)
        self.current_results = None
        self.error_msgs = []

    def __call__(self, real_obj, results):
        result = self.func(real_obj, results)
        self.current_results = results
        self._save_error_msgs(results)
        self.after_task_bp.stop()
        
        '''
        self.logger.info("task end, result:")
        if len(results) > 0:
            #self.logger.info(results[0].is_failed())
            #self.logger.info(dir(results[0]))
            if results[0].is_failed():
                self.logger.info(results[0]._result["stderr_lines"])
        #self.sync_obj.continue_runner_with_stop()
        '''

        return result


    def _save_error_msgs(self, results):
        for res in results:
            if res.is_failed() or res.is_unreachable():
                if hasattr(res, "_result"):
                    if "stderr" in res._result:
                        self.error_msgs.append(str(res._result["stderr"]))
                    elif "msg" in res._result:
                        self.error_msgs.append(str(res._result["msg"]))
                    elif "exception" in res._result:
                        self.error_msgs.append(str(res._result["exception"]))



import threading


class ans_sync:
    def __init__(self, logger):
        self.runner_event = threading.Event()
        self.ansible_event = threading.Event()
        self.logger = logger
        self.curr_breakpoint_label = None
        # Used to pass exceptions from Ansible thread
        self.exception = None

    def status(self):
        self.logger.debug("Runner event status: %s", self.runner_event.is_set())
        self.logger.debug("Ansible event status: %s",
                         self.ansible_event.is_set())

    def runner_just_wait(self):
        #self.logger.debug("runner: waiting...")
        self.runner_event.wait()
        self.runner_event.clear()
        if self.exception is not None:
            raise self.exception

    def ansible_just_wait(self):
        #self.logger.debug("ansible: waiting...")
        self.ansible_event.wait()
        self.ansible_event.clear()

    def continue_runner_with_stop(self, curr_breakpoint_label):
        #self.logger.debug("ansible: resume runner work and wait")
        self.curr_breakpoint_label = curr_breakpoint_label
        self.runner_event.set()
        self.ansible_event.wait()
        self.ansible_event.clear()

    def continue_ansible_with_stop(self):
        #self.logger.debug("runner: resume ansible work and wait")
        self.ansible_event.set()
        self.runner_event.wait()
        self.runner_event.clear()
        #self.logger.debug("runner: ANSIBLE WAKED ME UP")
        if self.exception is not None:
            raise self.exception

    def continue_runner(self):
        #self.logger.debug("ansible: resume runner work")
        self.runner_event.set()

    def continue_ansible(self):
        #self.logger.debug("runner: resume ansible work")
        self.ansible_event.set()



class ansi_breakpoint:
    def __init__(self, sync_obj, label):
        self.sync_obj = sync_obj
        self.label = label
    
    def stop(self):
        self.sync_obj.continue_runner_with_stop(self.label)


from cotea.wrappers.wrapper_base import wrapper_base
from cotea.wrappers.ansi_breakpoint import ansi_breakpoint


# wraps from ansible.cli.playbook.PlaybookCLI.run()
class pbcli_run_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger, before_bp_label, after_bp_label):
        super().__init__(func, sync_obj, logger)

        #self.parse_error = False
        self.before_playbook_bp = ansi_breakpoint(sync_obj, before_bp_label)
        self.after_playbook_bp = ansi_breakpoint(sync_obj, after_bp_label)


    # call of PlaybookCLI.run()
    def __call__(self, real_obj):
        self.logger.debug("before playbook")
        self.before_playbook_bp.stop()

        result = self.func(real_obj)

        self.logger.debug("after playbook")
        self.after_playbook_bp.stop()

        return result



from ansible.inventory.host import Host

from cotea.wrappers.wrapper_base import wrapper_base
from cotea.wrappers.ansi_breakpoint import ansi_breakpoint
from cotea.progress_bar import ansible_progress_bar

HOST_IND = 0
TASK_IND = 1


# wraps ansible.plugins.strategy.linear.StrategyModule._get_next_task_lockstep()
class get_next_task_wrapper(wrapper_base):
    def __init__(self, func, sync_obj, logger, bp_label, progress_bar: ansible_progress_bar, show_progr_bar=False):
        super().__init__(func, sync_obj, logger)
        self.before_task_bp = ansi_breakpoint(sync_obj, bp_label)
        self.prev_tasks = []
        self.next_tasks = []
        self.task_before_new = None
        self.show_progress_bar = show_progr_bar
        self.hosts_left = None

        self.rerun_last_task = False
        self.next_task_ignore_errors = False
        self.adding_last_task_after_new = True

        self.new_task_to_add = False
        self.play_iterator = None
        self.strategy_obj = None
        self.new_task = None

        self.should_ignored_errors_uuids = []
        self.should_ignored_unrch_uuids = []

        self.progress_bar = progress_bar


    def __call__(self, real_obj, hosts_left, iterator):
        result = None
        self.strategy_obj = real_obj
        self.play_iterator = iterator

        self.hosts_left = real_obj.get_hosts_left(iterator)

        if self.new_task_to_add:
            self.task_before_new = self._copy_hosttasks(self.prev_tasks)
            prev_tasks_with_new_task_set = self._set_task_in_hosttasks(self.new_task, self.task_before_new)
            self.next_tasks = prev_tasks_with_new_task_set
            result = prev_tasks_with_new_task_set

        elif self.rerun_last_task:
            if self.task_before_new:
                self.next_tasks = self._copy_hosttasks(self.task_before_new)
                self.task_before_new = None
            else:
                self.next_tasks = self._copy_hosttasks(self.prev_tasks)
            
            result = self.next_tasks

        else:
            result = self.func(real_obj, hosts_left, iterator)
            self.next_tasks = result

            self.should_ignored_errors_uuids = []
            self.should_ignored_unrch_uuids = []
                        
        for hosttask in result:
            if hosttask[TASK_IND]:
                # if we rerun the task or add the new one,
                # the list of already ignored tasks should stay the same
                if not self.rerun_last_task and not self.new_task_to_add:
                    if hosttask[TASK_IND].ignore_errors:
                        if hasattr(hosttask[TASK_IND], "get_name"):
                            if hasattr(hosttask[TASK_IND], "_uuid"):
                                self.should_ignored_errors_uuids.append(hosttask[TASK_IND]._uuid)
                    
                    if hosttask[TASK_IND].ignore_unreachable:
                        if hasattr(hosttask[TASK_IND], "get_name"):
                            if hasattr(hosttask[TASK_IND], "_uuid"):
                                self.should_ignored_unrch_uuids.append(hosttask[TASK_IND]._uuid)

                if self.next_task_ignore_errors:
                    hosttask[TASK_IND].ignore_errors = True
                    hosttask[TASK_IND].ignore_unreachable = True

        if self.get_next_task_name() is not None:
            self.progress_bar.update()
            
            if self.show_progress_bar:
                play_name = iterator._play.get_name()
                task_name = None
                if self.new_task_to_add:
                    if hasattr(self.new_task, "get_name"):
                        task_name = self.new_task.get_name()
                    else:
                        task_name = self.get_next_task_name()
                else:
                    task_name = self.get_next_task_name()
                    
                self.progress_bar.print_bar(play_name, task_name)
        
        if self.new_task_to_add and self.adding_last_task_after_new:
            self.rerun_last_task = True
        else:
            self.rerun_last_task = False

        self.next_task_ignore_errors = False
        self.new_task_to_add = False
        self.adding_last_task_after_new = True
        self.new_task = None

        self.before_task_bp.stop()

        return result


    def _copy_hosttasks(self, hosttasks):
        hosttasks_copy = []

        for hosttask in hosttasks:
            temp_host = None
            temp_task = None

            if hosttask[HOST_IND]:
                temp_host = Host()
                temp_host.deserialize(hosttask[HOST_IND].serialize())
            
            if hosttask[TASK_IND]:
                temp_task = hosttask[TASK_IND].copy()

            if temp_host:
                hosttasks_copy.append( (temp_host, temp_task) )
        
        return hosttasks_copy


    def set_next_to_prev(self):
        self.prev_tasks = self._copy_hosttasks(self.next_tasks)
        self.next_tasks = []


    def _set_task_in_hosttasks(self, new_task, hosttasks):
        hosttasks_with_new_task = []

        for hosttask in hosttasks:
            temp_host = None

            if hosttask[HOST_IND]:
                temp_host = Host()
                temp_host.deserialize(hosttask[HOST_IND].serialize())

            if temp_host:
                hosttasks_with_new_task.append( (temp_host, new_task) )
        
        return hosttasks_with_new_task


    def add_tasks(self, new_tasks):
        if self.strategy_obj and self.play_iterator:
            hosts_left = self.strategy_obj.get_hosts_left(self.play_iterator)

            for host in hosts_left:
                self.play_iterator.add_tasks(host, new_tasks)

            return True, ""
        
        error_msg = "Some of the needed objects are None. Most likely "
        error_msg += "the Ansible execution went further than expected"

        return False, error_msg


    def get_next_task(self):
        res = None
        
        if len(self.next_tasks) > 0:
            if len(self.next_tasks[0]) >= 2:
                res = self.next_tasks[0][TASK_IND]

        return res
    

    def get_next_task_name(self):
        res = None 
        next_task = self.get_next_task()

        if next_task:
            res = next_task.get_name()
        
        return res
    

    def get_prev_task(self):
        res = None

        if len(self.prev_tasks) > 0:
            if len(self.prev_tasks[0]) >= 2:
                res = self.prev_tasks[0][TASK_IND]

        return res
    
    
    def get_prev_task_name(self):
        res = None
        prev_task = self.get_prev_task()

        if prev_task:
            res = prev_task.get_name()
        
        return res
    
    
    def skip_next_task(self):
        skipped_tasks = []
        for host in self.hosts_left:
            host_state_block_and_task = self.play_iterator.get_next_task_for_host(host)
            skipped_task = host_state_block_and_task[len(host_state_block_and_task) - 1]
            skipped_tasks.append(skipped_task)

        skipped_task_name = ""

        if len(skipped_tasks) > 0:
            # task should be tha same on all hosts
            task = skipped_tasks[0]

            if hasattr(task, "get_name"):
                skipped_task_name = str(task.get_name())
        
        return skipped_task_name
    

    def dont_add_last_task_after_new(self):
        self.adding_last_task_after_new = False
